#
# vocalips010.tcl - GUI
# 
# Rev.0.10  Nov. 3, 2007  author: zhuo   www.geocities.jp/zhuoware
# 

set w .top
catch {destroy $w}
toplevel $w
wm title $w "VOCALIPS Rev.0.10"
wm iconname $w "VOCALIPS Rev.0.10"
wm withdraw .

proc createFileEntry { path text cmd } {
    frame $path
    label $path.lab -width 16 -text $text -anchor e
    entry $path.ent -width 80
    button $path.but -text "Browse ..." -command "$cmd $path"
    pack $path.lab -side left
    pack $path.ent -side left -expand yes -fill x
    pack $path.but -side left
    pack $path -fill x -padx 1c -pady 3
}

proc setEntryString { path string } {
    if {[string compare $string ""]} {
	$path delete 0 end
	$path insert 0 $string
	$path xview end
    }
}

createFileEntry $w.vsqfile "vocaloid file: " "vsqFileChooser $w.outfile.ent"
createFileEntry $w.outfile "output file: " "outFileChooser"
createFileEntry $w.tblfile "table file: " "tblFileChooser"

$w.tblfile.ent delete 0 end
$w.tblfile.ent insert 0 [lindex [glob -nocomplain -types f "*.tbl"] 0] 
$w.tblfile.ent xview end

frame $w.settings
labelframe $w.settings.mode -text "output mode" -padx 2 -pady 2
radiobutton $w.settings.mode.e0 -text "event-base, always with \"end\"" -variable outputMode -relief flat -value "e0" -anchor w
radiobutton $w.settings.mode.e1 -text "event-base, \"sil\" when early" -variable outputMode -relief flat -value "e1" -anchor w
radiobutton $w.settings.mode.f -text "frame-base" -variable outputMode -relief flat -value "f" -anchor w
pack $w.settings.mode.e0 $w.settings.mode.e1 $w.settings.mode.f -side top -pady 2 -fill x -expand no
pack $w.settings.mode -side left -expand yes -fill none
$w.settings.mode.f invoke

label $w.settings.fpslab -width 16 -text "fps" -anchor e
entry $w.settings.fpsent -width 8
$w.settings.fpsent insert 0 "30.00"
label $w.settings.offsetlab -width 16 -text "offset(msec)" -anchor e
entry $w.settings.offsetent -width 8
$w.settings.offsetent insert 0 "0.0"
pack $w.settings.fpslab $w.settings.fpsent $w.settings.offsetlab $w.settings.offsetent -side left -expand yes -fill none
pack $w.settings -side top -expand yes -fill none

frame $w.buttons
button $w.buttons.convert -width 16 -text "convert" -command "convert $w"
button $w.buttons.quit -width 16 -text quit -command "destroy $w; exit"
pack $w.buttons.convert -side left -expand 1
pack $w.buttons.quit -side left -expand 1
pack $w.buttons -side bottom -fill x -pady 5m

proc vsqFileChooser { out_ent path } {
    set types {{"vocaloid files" {.vsq}} {"All files" *}}
    set file [tk_getOpenFile -filetypes $types -parent $path -initialfile [$path.ent get]]
    regsub {\.vsq$} $file ".txt" outfile 
    setEntryString $path.ent $file
    setEntryString $out_ent $outfile
}

proc tblFileChooser { path } {
    set types {{"table files" {.tbl}} {"All files" *}}
    setEntryString $path.ent [tk_getOpenFile -filetypes $types -parent $path -initialfile [$path.ent get]]
}

proc outFileChooser { path } {
    set types {{"text files" {.txt}} {"All files" *}}
    setEntryString $path.ent [tk_getSaveFile -filetypes $types -parent $path -initialfile [$path.ent get]]
}

proc convert { w } {
    global outputMode
    set vsqfile [$w.vsqfile.ent get]
    set tblfile [$w.tblfile.ent get]
    set outfile [$w.outfile.ent get]
    set fps [$w.settings.fpsent get]
    set offset [$w.settings.offsetent get]
    if { $vsqfile == "" || ! [file exists $vsqfile] } {
	tk_messageBox -icon info -message "Please choose a vocaloid file." -type ok -parent $w
	return
    }
    if { $tblfile == "" || ! [file exists $tblfile] } {
	tk_messageBox -icon info -message "Please choose a table file." -type ok -parent $w
	return
    }
    if { $outfile == "" } {
	tk_messageBox -icon info -message "Please choose an output file." -type ok -parent $w
	return
    }
    if { $outputMode == "" } {
	tk_messageBox -icon info -message "Please choose an output mode." -type ok -parent $w
	return
    }
    file copy -force "$vsqfile" "T.VSQ"
    exec bin/mf2t.exe "T.VSQ" "T1.TXT"
    exec bin/Perl.exe bin/vsqtxt2lipsync_040.pl -m "$outputMode" -t "$tblfile" "$fps" "$offset" < "T1.TXT" > "T2.TXT"
    if { (![file exists "$outfile"]) || ([tk_messageBox -icon question -message "file exists: overwrite?  $outfile" -type okcancel -parent $w] == "ok") } {
	file copy -force "T2.TXT" "$outfile"
    }
    file delete "T.VSQ" "T1.TXT" "T2.TXT"
    tk_messageBox -icon info -message "$outfile created.\n" -type ok -parent $w
}
